import pyspark
from labelbox import Client as labelboxClient
from labelspark.uploader import create_upload_dict
from labelspark.connector import check_pyspark, get_col_names, get_unique_values
from labelbase.connector import validate_columns, determine_actions
from labelbase.uploader import create_global_key_to_data_row_id_dict, batch_create_data_rows, batch_rows_to_project, batch_upload_annotations
from labelbase.downloader import export_and_flatten_labels
import json
from delta import *
from uuid import uuid4
from labelbase.downloader import *
from labelbase.uploader import *
from labelbase.connector import *
from labelspark import connector, uploader

class Client:
    """ A Databricks Client, containing a Labelbox Client, to-be-run a Databricks notebook
    Args:
        lb_api_key                  :   Required (str) - Labelbox API Key
        lb_endpoint                 :   Optinoal (bool) - Labelbox GraphQL endpoint
        lb_enable_experimental      :   Optional (bool) - If `True` enables experimental Labelbox SDK features
        lb_app_url                  :   Optional (str) - Labelbox web app URL
    Attributes:
        lb_client                   :   labelbox.client.Client object
    Key Functions:
        create_data_rows_from_table :   Creates Labelbox data rows (and metadata) given a Databricks Spark table
        create_table_from_dataset   :   Creates a Databricks Spark table given a Labelbox dataset
        upsert_table_metadata       :   Updates Databricks Spark table metadata columns given a Labelbox dataset
        upsert_labelbox_metadata    :   Updates Labelbox metadata given a Databricks Spark table
    """
    def __init__(self, lb_api_key=None, lb_endpoint='https://api.labelbox.com/graphql', lb_enable_experimental=False, lb_app_url="https://app.labelbox.com"):
        self.lb_client = labelboxClient(lb_api_key, endpoint=lb_endpoint, enable_experimental=lb_enable_experimental, app_url=lb_app_url)
        check_pyspark()
        
    def export_to_table(
        self, project, 
        include_metadata:bool=False, include_performance:bool=False, include_agreement:bool=False,
        verbose:bool=False, mask_method:str="png", divider="///"):
        """ Creates a Spark DataFrame given a Labelbox Projet ID
        Args:
            project                 :   Required (str / lablebox.Project) - Labelbox Project ID or lablebox.Project object to export labels from
            include_metadata        :   Optional (bool) - If included, exports metadata fields
            include_performance     :   Optional (bool) - If included, exports labeling performance
            include_agreement       :   Optional (bool) - If included, exports consensus scores       
            verbose                 :   Optional (bool) - If True, prints details about code execution; if False, prints minimal information
            mask_method             :   Optional (str) - Specifies your desired mask data format
                                            - "url" leaves masks as-is
                                            - "array" converts URLs to numpy arrays
                                            - "png" converts URLs to png byte strings             
            divider                 :   Optional (str) - String delimiter for schema name keys and suffix added to duplocate global keys 
        """
        spark = pyspark.sql.SparkSession.builder.appName('labelspark_export').getOrCreate()  
        sc = spark.sparkContext
        
        flattened_labels_dict = export_and_flatten_labels(
            client=self.lb_client, project=project, 
            include_metadata=include_metadata, include_performance=include_performance, include_agreement=include_agreement,
            mask_method=mask_method, verbose=verbose, divider=divider
        )

        for label in flattened_labels_dict:
            for key in label.keys():
                if divider in key:
                    label[key] = json.dumps(label[key])

    
        df = spark.createDataFrame(sc.parallelize(flattened_labels_dict))
        
        if verbose:
            print(f"Success: DataFrame generated")
        
        return df        
    
    def export_to_delta_table(self, project, 
        include_metadata:bool=False, include_performance:bool=False, include_agreement:bool=False,
        verbose:bool=False, mask_method:str="png", divider="///", table_path="", write_mode="append", spark:pyspark.sql.SparkSession=None, spark_config:dict={}):
        """ Creates a Spark Dataframe and saves it to a Databricks Delta Table
        Args:
            project                 :   Required (str / lablebox.Project) - Labelbox Project ID or lablebox.Project object to export labels from
            include_metadata        :   Optional (bool) - If included, exports metadata fields
            include_performance     :   Optional (bool) - If included, exports labeling performance
            include_agreement       :   Optional (bool) - If included, exports consensus scores       
            verbose                 :   Optional (bool) - If True, prints details about code execution; if False, prints minimal information
            mask_method             :   Optional (str) - Specifies your desired mask data format
                                            - "url" leaves masks as-is
                                            - "array" converts URLs to numpy arrays
                                            - "png" converts URLs to png byte strings             
            divider                 :   Optional (str) - String delimiter for schema name keys and suffix added to duplicate global keys
            table_path              :   Required (str) - Path where the delta table will be saved
                                            - To save the delta table to a local folder, provide an absolute path
                                            - To save the delta table to a GCS bucket, provide a gs link with the format gs://{project}/{dataset}/{folder}
                                            - To save the delta table to an S3 bucket, provide an S3 link with the format s3a:/{bucket}/{folder}/{table_name}
            write_mode              :   Required (str) - used for writing Dataframe to Delta Table, must be either 'append' or 'overwrite'
            spark                   :   Optional (pyspark.sql.SparkSession) - SparkSession to use for exporting the project to a Delta Table, if not provided a spark session will
                                                                              be created based on the spark_config arg. This argument is mostly intended for debugging issues related to
                                                                              the required jar files for AWS/GCS connectors
            spark_config            :   Optional (dict) - spark_config needs to be in the following formats for uploading to AWS S3/GCS
                                            - S3: 
                                                {
                                                    "jars": [<path to hadoop-aws-{version}.jar file>, <path to aws-java-sdk-bundle-{version}jar file>],
                                                    "AWS_ACCESS_KEY": <AWS access key>,
                                                    "AWS_SECRET_KEY": <AWS secret key>
                                                }
                                                
                                            - GCS:
                                                {
                                                    "jars": [<path to gcs-connector-hadoop2-latest.jar file>],
                                                    "credentials": <path to GCS service account json keyfile>
                                                }"""

        if write_mode not in ['append', 'overwrite']:
            raise ValueError(f"The Delta Table write mode must be either 'append' or 'overwrite', the write mode provided is {write_mode}")
        
        if spark is None:
            spark = self.get_spark_session(table_path, spark_config)
        sc = spark.sparkContext
        
        flattened_labels_dict = export_and_flatten_labels(
            client=self.lb_client, project=project, 
            include_metadata=include_metadata, include_performance=include_performance, include_agreement=include_agreement,
            mask_method=mask_method, verbose=verbose, divider=divider
        )

        for label in flattened_labels_dict:
            for key in label.keys():
                if divider in key:
                    label[key] = json.dumps(label[key])
                elif label[key] == None:
                    label[key] = ""

        
        df = spark.createDataFrame(sc.parallelize(flattened_labels_dict))
        df.write.format("delta").mode(write_mode).save(table_path)
        
        if verbose:
            print(f"Success: Table saved to {table_path}")
        
        return df
    
    def get_spark_session(self, save_path, spark_config):
        if save_path[:5] == 'gs://':
            if 'credentials' not in spark_config.keys():
                raise ValueError(f"argument spark_config must contain a 'credentials' key to upload to GCS")
            if 'jars' not in spark_config.keys():
                raise ValueError(f"argument spark_config must contain a 'jars' key containing the path to the gcs-connector-hadoop2-latest.jar file to connect to GCS")
            if type(spark_config['jars']) == str:
                jar_file = spark_config['jars']
            elif type(spark_config['jars']) == list:
                jar_file = ''
                for file in spark_config['jars']:
                    jar_file += file + ', '
                jar_file = jar_file[:-2]
            builder = pyspark.sql.SparkSession.builder.appName("labelspark_export").config("spark.jars.packages", "io.delta:delta-core_2.12:2.2.0").config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension").config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog").config("spark.jars", jar_file)
            spark = configure_spark_with_delta_pip(builder).getOrCreate()
            spark.conf.set("spark.hadoop.fs.gs.auth.service.account.enable", "true")   
            spark.conf.set("google.cloud.auth.type", "SERVICE_ACCOUNT_JSON_KEYFILE")
            spark.conf.set('google.cloud.auth.service.account.json.keyfile', spark_config['credentials'])
            spark.conf.set("spark.hadoop.fs.gs.impl", "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFileSystem")
            spark._jsc.hadoopConfiguration().set("fs.gs.impl", "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFileSystem")
            spark.conf.set("fs.AbstractFileSystem.gs.impl", "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFileSystem")
            spark.conf.set("spark.delta.logStore.gs.impl", "io.delta.storage.GCSLogStore")
            spark.conf.set("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog")
            
        elif save_path[:5] == 's3a:/':
            if 'AWS_ACCESS_KEY' not in spark_config.keys() or 'AWS_SECRET_KEY' not in spark_config.keys():
                raise ValueError(f"argument spark_config must contain 'AWS_ACCESS_KEY' and 'AWS_SECRET_KEY' as keys to connect to S3")
            if 'jars' not in spark_config.keys():
                raise ValueError("argument spark_config must contain a 'jars' key with a list of strings containing the paths to hadoop-aws-{version}.jar and aws-java-sdk-bundle-{version}jar files")
            if type(spark_config['jars']) == str:
                jar_file = spark_config['jars']
            elif type(spark_config['jar']) == list:
                jar_file = ''
                for file in spark_config['jar']:
                    jar_file += file + ', '
                jar_file = jar_file[:-2]
            builder = pyspark.sql.SparkSession.builder.appName("labelspark_export").config("spark.jars.packages", "io.delta:delta-core_2.12:2.2.0").config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension").config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog").config("spark.jars", jar_file)
            spark = configure_spark_with_delta_pip(builder).getOrCreate()
            sc = spark.sparkContext
            hadoop_conf=sc._jsc.hadoopConfiguration()
            hadoop_conf.set("fs.s3a.impl", "org.apache.hadoop.fs.s3a.S3AFileSystem")
            hadoop_conf.set("fs.s3a.access.key", spark_config['AWS_ACCESS_KEY'])
            hadoop_conf.set("fs.s3a.secret.key", spark_config['AWS_SECRET_KEY'])
            sc._conf.setAll([('spark.delta.logStore.class','org.apache.spark.sql.delta.storage.S3SingleDriverLogStore')])
            spark.sparkContext._conf.getAll()

        else:
            builder = pyspark.sql.SparkSession.builder.appName("labelspark_export").config("spark.jars.packages", "io.delta:delta-core_2.12:2.2.0").config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension").config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog")
            spark = configure_spark_with_delta_pip(builder).getOrCreate()
        return spark
    
    def create_data_rows_from_table(
        self, table:pyspark.sql.dataframe.DataFrame, dataset_id:str="", project_id:str="", priority:int=5, 
        upload_method:str="", skip_duplicates:bool=False, model_id="", model_run_id="", mask_method:str="png", verbose:bool=False, divider="///"):
        """ Creates Labelbox data rows given a Spark DataFrame and a Labelbox Dataset
        Args:
            table               :   Required (pyspark.sql.dataframe.DataFrame) - Spark Table
            dataset_id          :   Required (str) - Labelbox dataset ID to add data rows to - only necessary if no "dataset_id" column exists            
            project_id          :   Required (str) - Labelbox project ID to add data rows to - only necessary if no "project_id" column exists
            priority            :   Optinoal (int) - Between 1 and 5, what priority to give to data row batches sent to projects                             
            upload_method       :   Optional (str) - Either "mal" or "import" - required to upload annotations (otherwise leave as "")
            skip_duplicates     :   Optional (bool) - Determines how to handle if a global key to-be-uploaded is already in use
                                        If True, will skip duplicate global_keys and not upload them
                                        If False, will generate a unique global_key with a suffix {divider} + "1", "2" and so on
            mask_method         :   Optional (str) - Specifies your input mask data format
                                        - "url" means your mask is an accessible URL (must provide color)
                                        - "array" means your mask is a numpy array (must provide color)
                                        - "png" means your mask value is a png-string                                           
            verbose             :   Optional (bool) - If True, prints details about code execution; if False, prints minimal information               
            divider             :   Optional (str) - String delimiter for schema name keys and suffix added to duplocate global keys
        """
        # Create/identify the following values:
            # row_data_col      : column with name "row_data"
            # global_key_col    : column with name "global_key" - defaults to row_data_col
            # external_id_col   : column with name "external_id" - defaults to global_key_col
            # project_id_col    : column with name "project_id" - defaults to ""
            # dataset_id_col    : column with name "dataset_id" - defaults to ""
            # metadata_index    : Dictonary where {key=metadata_field_name : value=metadata_type} - defaults to {}
            # attachment_index  : Dictonary where {key=column_name : value=attachment_type} - defaults to {}
            # annotation_index  : Dictonary where {key=column_name : value=top_level_feature_name} - defaults to {}
        x = validate_columns(
            client=self.lb_client, table=table,
            get_columns_function=get_col_names,
            get_unique_values_function=get_unique_values,
            divider=divider, verbose=verbose, extra_client=None
        )
                
        # Determine if we're batching and/or uploading annotations
        actions = determine_actions(
            row_data_col=x["row_data_col"], dataset_id=dataset_id, dataset_id_col=x["dataset_id_col"], project_id=project_id, 
            project_id_col=x["project_id_col"], model_id=model_id, model_id_col=x["model_id_col"], model_run_id=model_run_id, 
            model_run_id_col=x["model_run_id_col"], upload_method=upload_method, metadata_index=x["metadata_index"], 
            attachment_index=x["attachment_index"], annotation_index=x["annotation_index"], prediction_index=x["prediction_index"]
        )
        if not actions["create"]:
            raise ValueError(f"No `dataset_id` argument or `dataset_id` column was provided in table to create data rows")
        
        # Create an upload dictionary where {
            # dataset_id : {
                # global_key : {
                    # "data_row" : {}, -- This is your data row upload as a dictionary
                    # "project_id" : "", -- This batches data rows to projects, if applicable
                    # "annotations" : [] -- List of annotations for a given data row, if applicable
                # }
            # }
        # }
        # This uniforms the upload to use labelbase - Labelbox base code for best practices
        upload_dict = create_upload_dict(
            client=self.lb_client, table=table, 
            row_data_col=x["row_data_col"], global_key_col=x["global_key_col"], external_id_col=x["external_id_col"], 
            dataset_id_col=x["dataset_id_col"], dataset_id=dataset_id, 
            project_id_col=x["project_id_col"], project_id=project_id,
            metadata_index=x["metadata_index"], attachment_index=x["attachment_index"], 
            annotation_index=x["annotation_index"],
            upload_method=upload_method, mask_method=mask_method, divider=divider, verbose=verbose
        )    
                
        # Upload your data rows to Labelbox - update upload_dict if global keys are modified during upload
        data_row_upload_results, upload_dict = batch_create_data_rows(
            client=self.lb_client, upload_dict=upload_dict, 
            skip_duplicates=skip_duplicates, divider=divider, verbose=verbose
        )
        
        # Bath to project attempt
        if actions['batch']: 
            try:
                # Create a dictionary where {key=global_key : value=data_row_id}
                global_keys_list = []
                for dataset_id in upload_dict.keys():
                    for global_key in upload_dict[dataset_id].keys():
                        global_keys_list.append(global_key)
                global_key_to_data_row_id = create_global_key_to_data_row_id_dict(
                    client=self.lb_client, global_keys=global_keys_list
                )     
                # Create batch dictionary where {key=project_id : value=[data_row_ids]}   
                project_id_to_batch_dict = {}                
                for dataset_id in upload_dict:
                    for global_key in upload_dict[dataset_id].keys():                    
                        project_id = upload_dict[dataset_id][global_key]["project_id"]
                        if project_id:
                            if project_id not in project_id_to_batch_dict.keys():
                                project_id_to_batch_dict[project_id] = []
                            project_id_to_batch_dict[project_id].append(global_key_to_data_row_id[global_key])
                # Labelbase command to batch data rows to projects
                batch_to_project_results = batch_rows_to_project(
                    client=self.lb_client, project_id_to_batch_dict=project_id_to_batch_dict, 
                    priority=priority, verbose=verbose
                )
            except Exception as e:
                actions['annotate'] = False
                batch_to_project_results = e
        else:
            batch_to_project_results = []                
        
        # Annotation upload attempt
        if actions['annotate']:
            try:               
                # Create batch dictionary where {key=project_id : value=[data_row_ids]}
                project_id_to_upload_dict = {}
                for dataset_id in upload_dict.keys():
                    for global_key in upload_dict[dataset_id].keys():
                        project_id = upload_dict[dataset_id][global_key]["project_id"]
                        annotations = []
                        annotations_no_data_row_id = upload_dict[dataset_id][global_key]["annotations"]
                        for annotation in annotations_no_data_row_id:
                            for key in annotation.keys():
                                annotations.append(annotation[key])
                        annotations_no_data_row_id = annotations
                        if project_id not in project_id_to_upload_dict.keys():
                            project_id_to_upload_dict[project_id] = []
                        if annotations_no_data_row_id: # For each annotation in the list, add the data row ID and add it to your annotation upload dict
                            for annotation in annotations_no_data_row_id:
                                annotation_data_row_id = annotation
                                annotation_data_row_id["dataRow"] = {"id" : global_key_to_data_row_id[global_key]}
                                project_id_to_upload_dict[project_id].append(annotation_data_row_id)
                # Labelbase command to upload annotations to projects
                annotation_upload_results = batch_upload_annotations(
                    client=self.lb_client, project_id_to_upload_dict=project_id_to_upload_dict, how=upload_method, verbose=verbose
                )
            except Exception as e:
                annotation_upload_results = e
        else:
            annotation_upload_results = []          
            
        return {
            "data_row_upload_results" : data_row_upload_results, 
            "batch_to_project_results" : batch_to_project_results,
            "annotation_upload_results" : annotation_upload_results
        }
    
    def create_data_rows_from_delta_table(self, dataset_id:str="", project_id:str="", priority:int=5, 
        upload_method:str="", skip_duplicates:bool=False, mask_method:str="png", verbose:bool=False, divider="///", table_path="", spark:pyspark.sql.SparkSession=None, spark_config:dict={}):
        
        if spark is None:
            spark = self.get_spark_session(table_path, spark_config)

        df = spark.read.load(table_path)
        return(self.create_data_rows_from_table(table=df, dataset_id=dataset_id, project_id=project_id, priority=priority, upload_method=upload_method, skip_duplicates=skip_duplicates, mask_method=mask_method, verbose=verbose, divider=divider))

    def upsert_data_rows_from_table(self, table:pyspark.sql.dataframe.DataFrame, dataset_id:str="", project_id:str="", model_id:str="", upload_method:str="", mask_method:str="png", priority:int=5, model_run_id:str="", batch_data_rows:bool=False, annotation_method:str="", verbose:bool=False, divider:str="///"):
        """ Performs the following actions if proper information is provided:
                - batches data rows to projects (if batch_data_rows == True) * **
                - uploads annotations as pre-labels or submitted labels * **
                - sends submitted labels to model runs as ground truth labels * ** ***
                - uploads predictions to a model run * ***
            * Requires a global_key AND dataset_id OR dataset_id_col
            ** Requires project_id OR project_id_col
            *** Requires model_id OR model_id_col OR model_run_id OR model_run_id_col      
        Args:
            batch_data_rows     :   Optional (bool) - If True, will batch data rows to projects if not already in projects
            anootation_method   :   Optional (bool) -  Must be one of the following annotation import methods:
                                            "" - Does nothing with annotations
                                            "mal" - Uploaded to project as pre-labels
                                            "import" - Uploaded to project assubmitted labels
                                            "ground-truth" - Project submitted labels AND/OR ground truth submitted labels (will check existing project labels first)
            upsert_metadata     :   True if upserting metadata, False if not
            upsert_attachments  :   True if upserting attachments, False if not
        Returns:
            Results from all performed actions in a dictionary - if an expected action has no results, it was not performed
        """
        # Confirm the existience of specific columns and construct reference indexes for different kinds of data to-be-uploaded using labelbase.connector.validate_columns
            # row_data_col                :   Column representing asset URL, raw text, or path to local asset file
            # global_key_col              :   Defaults to row_data_col
            # external_id_col             :   Defaults to global_key_col
            # project_id_col              :   Either `project_id` or "" - overridden by arg for project_id        
            # dataset_id_col              :   Either `dataset_id` or "" - overridden by arg for dataset_id
            # model_id_col                :   Either `model_id` or "" - overridden by arg for model_id and any args for model_run_id
            # model_run_id_col            :   Either `model_run_id` or "" - overridden by arg for model_run_id        
            # metadata_index              :   Dictonary where {key=metadata_field_name : value=metadata_type} or {} if not uploading metadata
            # attachment_index            :   Dictonary where {key=column_name : value=attachment_type} or {} if not uploading attachments
            # annotation_index            :   Dictonary where {key=column_name : value=top_level_class_name} or {} if not uploading annotations
            # prediction_index            :   Dictonary where {key=column_name : value=top_level_class_name} or {}  if not uploading predictions
        
        x = validate_columns(
            client=self.lb_client, table=table,
            get_columns_function=connector.get_col_names,
            get_unique_values_function=connector.get_unique_values,
            divider=divider, verbose=verbose, extra_client=None
        )
        
        # Determine the actions we're taking in this run using labelbase.connector.determine_actions
        actions = determine_actions(
            row_data_col=x["row_data_col"], dataset_id=dataset_id, dataset_id_col=x["dataset_id_col"], project_id=project_id, 
            project_id_col=x["project_id_col"], model_id=model_id, model_id_col=x["model_id_col"], model_run_id=model_run_id, 
            model_run_id_col=x["model_run_id_col"], upload_method=upload_method, metadata_index=x["metadata_index"], 
            attachment_index=x["attachment_index"], annotation_index=x["annotation_index"], prediction_index=x["prediction_index"]
        )             
        
        # Create an upload dictionary where:
        # {
            # global_key : {
                # "project_id" : "", -- This batches data rows to projects, if applicable
                # "annotations" : [], -- List of annotations for a given data row, if applicable
                # "model_run_id" : "", -- This adds data rows to model runs, if applicable
                # "predictions" : [] -- List of predictions for a given data row, if applicable
            # }
        # }
        # This uniforms the upload to use labelbase - Labelbox base code for best practices
        upload_dict = uploader.create_upload_dict( # Using labelspark.uploader.create_upload_dict
            client=self.lb_client, table=table, 
            row_data_col=x["row_data_col"], global_key_col=x["global_key_col"], external_id_col=x["external_id_col"], 
            dataset_id_col=x["dataset_id_col"], dataset_id=dataset_id, 
            project_id_col=x["project_id_col"], project_id=project_id,
            metadata_index=x["metadata_index"], attachment_index=x["attachment_index"], 
            annotation_index=x["annotation_index"], upload_method=upload_method, 
            mask_method=mask_method, divider=divider, verbose=verbose
        )
        for gk in upload_dict.keys():
            annotations_filtered = []
            annotations = upload_dict[gk]['annotations']
            for annotation in annotations:
                for key in annotation.keys():
                    annotations_filtered.append(annotation[key])
            upload_dict[gk]['annotations'] = annotations_filtered

        # Bath to project attempt using labelbase.uploader.batch_rows_to_project
        if actions["batch"] and batch_data_rows:         
            batch_to_project_results = batch_rows_to_project(
                client=self.lb_client, upload_dict=upload_dict, priority=priority, verbose=verbose
            )                
        else:
            batch_to_project_results = []

        # Annotation upload attempt using labelbase.uploader.batch_upload_annotations
        if actions["annotate"]:
            annotation_upload_results = batch_upload_annotations(
                client=self.lb_client, upload_dict=upload_dict, how=actions["annotate"], verbose=verbose
            )
        else:
            annotation_upload_results = []

        # If model run upload attempt labelbase.uploader.batch_add_ground_truth_to_model_run
        if actions["annotate"] == "ground-truth":
            ground_truth_upload_results = batch_add_ground_truth_to_model_run(
                client=self.lb_client, upload_dict=upload_dict
            )
        else:
            ground_truth_upload_results = []
            
        # Prediction upload attempt using labelbase.uploader.batch_upload_predictions
        if actions["predictions"]:
            # If data rows aren't in the model run yet, add them with labelbase.uploader.batch_add_data_rows_to_model_run
            if not ground_truth_upload_results:
                data_row_to_model_run_results = batch_add_data_rows_to_model_run(
                    client=self.lb_client, upload_dict=upload_dict                    
                )
            prediction_upload_results = batch_upload_predictions(
                client=self.lb_client, upload_dict=upload_dict
            )
        else:
            data_row_to_model_run_results = []
            prediction_upload_results = []  

        return_payload = {
            "batch_to_project_results" : batch_to_project_results,
            "annotation_upload_results" : annotation_upload_results,
            "ground_truth_upload_results" : ground_truth_upload_results,
            "data_row_to_model_run_results" : data_row_to_model_run_results,
            "prediction_upload_results" : prediction_upload_results
        }             

        return_payload = {k: v for k, v in return_payload.items() if v}
            
        return return_payload

    def upsert_data_rows_from_delta_table(self, dataset_id:str="", project_id:str="", model_id:str="", upload_method:str="", mask_method:str="png", priority:int=5, model_run_id:str="", batch_data_rows:bool=False, annotation_method:str="", verbose:bool=False, divider:str="///", table_path:str="", spark:pyspark.sql.SparkSession=None, spark_config:dict={}):
        """ Performs the following actions if proper information is provided:
                - batches data rows to projects (if batch_data_rows == True) * **
                - uploads annotations as pre-labels or submitted labels * **
                - sends submitted labels to model runs as ground truth labels * ** ***
                - uploads predictions to a model run * ***
            * Requires a global_key AND dataset_id OR dataset_id_col
            ** Requires project_id OR project_id_col
            *** Requires model_id OR model_id_col OR model_run_id OR model_run_id_col      
        Args:
            batch_data_rows     :   Optional (bool) - If True, will batch data rows to projects if not already in projects
            anootation_method   :   Optional (bool) -  Must be one of the following annotation import methods:
                                            "" - Does nothing with annotations
                                            "mal" - Uploaded to project as pre-labels
                                            "import" - Uploaded to project assubmitted labels
                                            "ground-truth" - Project submitted labels AND/OR ground truth submitted labels (will check existing project labels first)
            upsert_metadata     :   True if upserting metadata, False if not
            upsert_attachments  :   True if upserting attachments, False if not
        Returns:
            Results from all performed actions in a dictionary - if an expected action has no results, it was not performed
        """
        if spark is None:
            spark = self.get_spark_session(table_path, spark_config)

        df = spark.read.load(table_path)
        return self.upsert_data_rows_from_table(df, dataset_id, project_id, model_id, upload_method, mask_method, priority, model_run_id, batch_data_rows, annotation_method, verbose, divider)


"""
uploader.py holds the function create_upload_dict() -- which creates the following style dictionary:
{
    dataset_id : {
        global_key : {
            "data_row" : {}, -- This is your data row upload as a dictionary
            "project_id" : "", -- This batches data rows to projects, if applicable
            "annotations" : [] -- List of annotations for a given data row, if applicable -- note this does not contain required data row ID information
        }
        global_key : {
            "data_row" : {},
            "project_id" : "",
            "annotations" : []
        }
    },
    dataset_id : {
        global_key : {
            "data_row" : {},
            "project_id" : "",
            "annotations" : []
        }
        global_key : {
            "data_row" : {},
            "project_id" : "",
            "annotations" : []
        }
    }
}

This uniforms input formats so that they can leverage labelbase - Labelbox base code for best practices
"""
import pyspark
from pyspark.sql.functions import lit, udf
from pyspark.sql.types import StructType, StructField, StringType, MapType, ArrayType
from labelbox import Client as labelboxClient
from labelbox.schema.ontology import Ontology as labelboxOntology
from labelspark.connector import get_table_length, get_unique_values
from labelbase.metadata import get_metadata_schema_to_name_key, process_metadata_value
from labelbase.ontology import get_ontology_schema_to_name_path
from labelbase.annotate import create_ndjsons
import json
from concurrent.futures import ThreadPoolExecutor, as_completed

def create_upload_dict(client:labelboxClient, table:pyspark.sql.dataframe.DataFrame,
                       row_data_col:str, global_key_col:str, external_id_col:str, 
                       dataset_id_col:str, dataset_id:str,
                       project_id_col:str, project_id:str,
                       metadata_index:dict, attachment_index:dict, annotation_index:dict,
                       upload_method:str, mask_method:str, divider:str, verbose:bool, extra_client:bool=None):  
    """ Uses UDFs to create a column of data row dictionaries to-be-uploaded, then converts this column into a list
    Args:
        client                      :   Required (labelbox.client.Client) - Labelbox Client object        
        table                       :   Required (pandas.core.frame.DataFrame) - Pandas DataFrame                
        row_data_col                :   Required (str) - Column containing asset URL or raw text
        global_key_col              :   Required (str) - Column name containing the data row global key - defaults to row data
        external_id_col             :   Required (str) - Column name containing the data row external ID - defaults to global key
        dataset_id_col              :   Required (str) - Column name containing the dataset ID to add data rows to
        dataset_id                  :   Required (str) - Labelbox dataset ID to add data rows to - only necessary if no "dataset_id" column exists            
        project_id_col              :   Required (str) - Column name containing the project ID to batch a given row to        
        project_id                  :   Required (str) - Labelbox project ID to add data rows to - only necessary if no "project_id" column exists
        metadata_index              :   Required (dict) - Dictonary where {key=metadata_field_name : value=metadata_type}
        attachment_index            :   Required (dict) - Dictonary where {key=column_name : value=attachment_type}
        annotation_index            :   Required (dict) - Dictonary where {key=column_name : value=top_level_feature_name}
        upload_method               :   Required (str) - Either "mal" or "import" - required to upload annotations (otherwise leave as "")
        mask_method                 :   Optional (str) - Specifies your input mask data format
                                            - "url" means your mask is an accessible URL (must provide color)
                                            - "array" means your mask is a numpy array (must provide color)
                                            - "png" means your mask value is a png-string                 
        divider                     :   Required (str) - String delimiter for all name keys generated
        verbose                     :   Required (bool) - If True, prints details about code execution; if False, prints minimal information
        extra_client                :   Ignore this value - necessary for other labelbase integrations                
    Returns:
        - global_key_to_upload_dict - Dictionary in the above format
    """     
    # Check that global key column is entirely unique values
    table_length = get_table_length(table=table, extra_client=extra_client)
    if verbose:
        print(f'Creating upload list - {table_length} rows in Pandas DataFrame')
    unique_global_key_count = len(get_unique_values(table=table, col=global_key_col, extra_client=extra_client))
    if table_length != unique_global_key_count:
        print(f"Warning: Your global key column is not unique - upload will resume, only uploading 1 data row per unique global key")      

    # Create your column of upload dict values using UDFs
    uploads_table = create_uploads_column(
        client=client, table=table, row_data_col=row_data_col, global_key_col=global_key_col, external_id_col=external_id_col, 
        dataset_id_col=dataset_id_col, dataset_id=dataset_id, project_id_col=project_id_col, project_id=project_id, metadata_index=metadata_index, 
        attachment_index=attachment_index, annotation_index=annotation_index, 
        upload_method=upload_method, mask_method=mask_method, divider=divider, verbose=verbose 
    )
    # Query your uploads column and create your upload dict

    upload_dict = {}
    res = uploads_table.select("uploads").rdd.map(lambda x: {'data_row': {'row_data': x['uploads']['data_row']['row_data'], 'global_key': x['uploads']['data_row']['global_key'], 'external_id': x['uploads']['data_row']['external_id'], 'metadata_fields': x['uploads']['data_row']['metadata_fields'], 'attachments': x['uploads']['data_row']['attachments']}, 'dataset_id': x['uploads']['dataset_id'], 'project_id': x['uploads']['project_id'], 'annotations': x['uploads']['annotations']}).collect()

    with ThreadPoolExecutor() as exc:
        futures = [exc.submit(process_upload, x) for x in res]
        for future in as_completed(futures):
            y = future.result()
            upload_dict[y["data_row"]["global_key"]] = {
                "data_row" : y["data_row"],
                "project_id" : y["project_id"],
                "annotations" : y["annotations"],
                "dataset_id" : y["dataset_id"]
            }        
    return upload_dict

def create_uploads_column(client:labelboxClient, table:pyspark.sql.dataframe.DataFrame,
                          row_data_col:str, global_key_col:str, external_id_col:str, 
                          dataset_id_col:str, dataset_id:str, project_id_col:str, project_id:str,
                          metadata_index:dict, attachment_index:dict, annotation_index:dict,
                          upload_method:str, mask_method:str, divider:str, verbose:bool, extra_client:bool=None):
    """ Takes a table and returns a new table with an "uploads" column where the values are
    {
        "data_row" : {
            "global_key" : "",                    |
            "row_data" : "",                      |
            "external_id" : "",                   | ---- This is your data row upload as a dictionary
            "metadata_fields" : [],               |
            "attachments" : []                    |
        }, 
        "dataset_id" : "" -- This is the dataset ID to upload this data row to
        "project_id" : "", -- This batches data rows to projects, if applicable
        "annotations" : [] -- List of annotations for a given data row, if applicable (note that data row ID is not included)
    }
    """
    # Get dictionary where {key=project_id : value=ontology_index} -- index created by labelbase -- if project IDs are available
    if project_id != "":
        project_ids = [project_id]
    elif project_id_col != "":
        project_ids = get_unique_values(table=table, col=project_id_col, extra_client=extra_client)
    else:
        project_ids = []
    project_id_to_ontology_index = {}        
    if (project_ids!=[]) and (annotation_index!={}) and (upload_method in ["mal", "import"]): # If we're uploading annotations
        for projectId in project_ids:
            ontology = client.get_project(projectId).ontology()
            project_type = client.get_project(projectId).media_type
            project_id_to_ontology_index[projectId] = get_ontology_schema_to_name_path(ontology=ontology,divider=divider,invert=True,detailed=True)
            project_id_to_ontology_index[projectId]['project_type'] = str(project_type)

    project_id_to_ontology_index_bytes = json.dumps(project_id_to_ontology_index)          
    # Create your upload column's syntax
    upload_schema = StructType([
        StructField(
            "data_row", StructType([
                StructField("row_data", StringType()), StructField("global_key", StringType()), StructField("external_id", StringType()),
                StructField("metadata_fields", ArrayType(MapType(StringType(), StringType(), True))),
                StructField("attachments", ArrayType(MapType(StringType(), StringType(), True)))
            ])
        ),
        StructField("dataset_id", StringType()), StructField("project_id", StringType()), 
        StructField("annotations", ArrayType(MapType(StringType(), StringType(), True)))
    ])
    x = get_metadata_schema_to_name_key(client=client, divider=divider, invert=True) # Get metadata dict where {key=name : value=schema_id}
    metadata_name_key_to_schema_bytes = json.dumps(x) # Convert reference dict to bytes    
    # Run a UDF to create row values
    data_rows_udf = udf(create_data_row, upload_schema)    
    project_input = lit(project_id_col) if not project_id_col else project_id_col
    dataset_input = lit(dataset_id_col) if not dataset_id_col else dataset_id_col

    table = table.withColumn(
      'uploads', data_rows_udf(
          row_data_col, global_key_col, external_id_col, lit(metadata_name_key_to_schema_bytes),
          lit(project_id_col), project_input, lit(project_id), lit(dataset_id_col), dataset_input, lit(dataset_id)
      )
    )
    print(table.collect()[0])
    # Run a UDF to add attachments, if applicable  
    if attachment_index:
        print(attachment_index)
        attachments_udf = udf(create_attachments, upload_schema)  # Create a UDF
        for attachment_column_name in attachment_index: # Run this UDF for each attachment column in the attachment index
            attachment_type = attachment_index[attachment_column_name]
            table = table.withColumn('uploads', attachments_udf('uploads', lit(attachment_type), attachment_column_name))        
    # Run a UDF to add metadata, if applicable
    if metadata_index:
        print(metadata_index)
        metadata_udf = udf(create_metadata, upload_schema) # Create a UDF
        for metadata_field_name in metadata_index: # Run this UDF for each metadata field name in the metadata index
            metadata_type = metadata_index[metadata_field_name]
            metadata_column_name = f"metadata{divider}{metadata_type}{divider}{metadata_field_name}"
            table = table.withColumn(
                'uploads', metadata_udf(
                    'uploads', lit(metadata_field_name), metadata_column_name, lit(metadata_type), lit(metadata_name_key_to_schema_bytes), lit(divider)
                )
            )    
    # Run a UDF to add annotations, if applicable 
    if (annotation_index!={}) and (project_id_to_ontology_index!={}) and (upload_method in ["mal", "import"]):
        annotation_udf = udf(create_annotations, upload_schema) # Create a UDF
        for annotation_column_name in annotation_index: # Run this UDF for each attachment column name in the attachment index
            top_level_feature_name = annotation_index[annotation_column_name]
            annotation_type = annotation_column_name.split(divider)[1]
            table = table.withColumn(
              'uploads', annotation_udf(
                  'uploads', lit(top_level_feature_name), annotation_column_name, lit(mask_method), lit(annotation_type), lit(project_id_to_ontology_index_bytes), lit(divider)
              )
            )        
    return table

def create_data_row(row_data_col, global_key_col, external_id_col, metadata_name_key_to_schema_bytes,
                    project_id_col_name, project_id_col_value, project_id_str,
                    dataset_id_col_name, dataset_id_col_value, dataset_id_str):
    """ Function to-be-wrapped in a UDF that creates upload dict values (without metadata, attachments or annotations)
    """
    metadata_name_key_to_schema = json.loads(metadata_name_key_to_schema_bytes)
    if project_id_col_name:
        projectId = project_id_col_value
    elif project_id_str:
        projectId = project_id_str
    else:
        projectId = ""    
    if dataset_id_col_name:
        datasetId = dataset_id_col_value
    else:
        datasetId = dataset_id_str        
    data_row_dict = {
        "row_data" : row_data_col, "external_id" : external_id_col, "global_key" : global_key_col,
        "metadata_fields" : [{"schema_id" : metadata_name_key_to_schema["lb_integration_source"], "value" : "Databricks"}],
        "attachments" : []
    }
    return { "data_row" : data_row_dict, "project_id" : projectId, "dataset_id" : datasetId, "annotations" : [] }

def create_metadata(uploads_col, metadata_field_name, metadata_col_value, metadata_type, metadata_name_key_to_schema_bytes, divider):
    """ Function to-be-wrapped in a UDF that adds metadata fields to your upload dict
    """  
    metadata_name_key_to_schema = json.loads(metadata_name_key_to_schema_bytes)
    input_metadata = process_metadata_value(
        metadata_value=metadata_col_value, metadata_type=metadata_type, parent_name=metadata_field_name, 
        metadata_name_key_to_schema=metadata_name_key_to_schema, divider=divider
    )            
    if input_metadata:
        uploads_col["data_row"]["metadata_fields"].append({"schema_id" : metadata_name_key_to_schema[metadata_field_name], "value" : input_metadata})
    return uploads_col

def create_attachments(uploads_col, attachment_type, attachment_col_value):
    """ Function to-be-wrapped in a UDF that adds attachments to your upload dict
    """  
    if attachment_col_value:
        uploads_col["data_row"]["attachments"].append({"type" : attachment_type, "value" : attachment_col_value})
    return uploads_col  

def create_annotations(uploads_col, top_level_feature_name, annotations, mask_method, annotation_type, project_id_to_ontology_index_bytes, divider):
    """ Function to-be-wrapped in a UDF that adds attachments to your upload dict
    """  
    project_id_to_ontology_index = json.loads(project_id_to_ontology_index_bytes)
    ontology_index = project_id_to_ontology_index[uploads_col["project_id"]]
    print(annotations)
    if annotations is not None:
        annotation_list = []
        ndjsons = create_ndjsons(
                    top_level_name=top_level_feature_name,
                    annotation_inputs=annotations,
                    ontology_index=ontology_index,
                    mask_method=mask_method,
                    divider=divider    
                )
        for ndjson in ndjsons:
            annotation_list.append({annotation_type:json.dumps(ndjson)})
        print(annotation_list)
        uploads_col["annotations"].extend(
            annotation_list
        )
    return uploads_col


def process_upload(x):
    """ Function to-be-multithreaded that processes rows into the upload_dict format
    """
    annotations = [string_to_ndjson(x) for x in x["annotations"]]
    return {
        "data_row" : x["data_row"],
        "project_id" : x["project_id"],
        "annotations" : annotations,
        "dataset_id" : x["dataset_id"]
    }

def string_to_ndjson(annotation):
    """ Function that takes stringtype representation of an annotation and converts it into the Labelbox ndjson format
    """
    for key, value in annotation.items():
        annotation[key] = json.loads(value)
    return annotation
    



from pyspark.sql.functions import udf, lit
from pyspark.sql.types import StringType
import json

def update_metadata(client, spark_dataframe, metadata_field_name, lb_dataset):
    """ Updates a spark dataframe with the current Labelbox metadata value given a metadata field name
    Args:
        client                      :       labelbox.Client object
        spark_dataframe             :       pyspark.sql.dataframe.Dataframe object - must have "data_row_id" and metadata_field_name columns at a minimum
        metadata_field_name         :       Name of the column in the spark_dataframe to-be-updated. This must also be the name of the metadata field in Labelbox
        lb_dataset                  :       labelbox.schema.dataset.Dataset object that hosts the data row IDs in question
    Returns:
        Updated spark_dataframe
    """

    mdo = client.get_data_row_metadata_ontology()

    data_row_ids = [data_row.uid for data_row in lb_dataset.export_data_rows()]
    data_row_metadata = mdo.bulk_export(data_row_ids)

    mdo_reserved_by = mdo.reserved_by_name
    mdo_reserved_by.update(mdo.custom_by_name)
    metadata_field_schema_id = mdo_reserved_by[metadata_field_name].uid

    update_dict = {}
    for data_row in data_row_metadata:
        key = data_row.data_row_id
        for field in data_row.fields:
            if field.schema_id == metadata_field_schema_id:
                value = field.value
                update_dict[key] = value  

    def sync_function(update_dict, data_row_id, current_metadata_value):
        """ Nested UDF Functionality update metadata in Labelbox and in the pyspark dataframe
        Args:  
            update_dict                 :      Dictionary where {key=data_row_id and value=latest_metadata_value}
            data_row_id                 :      Labelbox Data Row ID
            current_metadata_value      :      The existing value for this metadata field in the spark_dataframe
        Returns:
            New value to-be-inserted in the column corresponding to this metadata field
        """  
        reference_dict = json.loads(update_dict)
        if data_row_id in reference_dict.keys():
            return_value = reference_dict[data_row_id]
        else:
            return_value = current_metadata_value
        return return_value                
                
    sync_udf = udf(sync_function, StringType())

    return spark_dataframe.withColumn(metadata_field_name, sync_udf(lit(json.dumps(update_dict)), "data_row_id", metadata_field_name))



"""
connector.py holds the following helper functions specific to pandas DataFrames:

- get_col_names(table)                                    : Gets all column names
- get_unique_values(table, col)                           : Gets all unique values in a given column
- add_col(table, col, default)                            : Creates a column where all values equal the default value
- get_table_length(table)                                 : Gets the length of a DataFrame
- rename_col(table, col, to)                              : Renames a given column

"""

import pyspark

def check_pyspark():
    try:
        import pyspark.pandas as pd
    except:
        raise RuntimeError(f'labelspark.Client() requires pyspark to be installed - please update your Databricks runtime to support pyspark')
    
def get_col_names(table:pyspark.sql.dataframe.DataFrame, extra_client=None):
    """Grabs all column names from a Spark Dataframe
    Args:
        table           :   Required (pyspark.sql.dataframe.DataFrame) - Spark Table
        extra_client    :   Ignore this value - necessary for other labelbase integrations
    Returns:
        List of strings corresponding to all column names
    """
    return [str(col[0]) for col in table.dtypes]    

def get_unique_values(table:pyspark.sql.dataframe.DataFrame, col:str, extra_client=None):
    """ Grabs all unique values from a Spark Dataframe column as strings
    Args:
        table           :   Required (pyspark.sql.dataframe.DataFrame) - Spark Table
        col             :   Required (str) - Spark Table column name
        extra_client    :   Ignore this value - necessary for other labelbase integrations
    Returns:
        List of unique values from a spark table column as strings
    """
    return [str(x.__getitem__(col)) for x in table.select(col).distinct().collect()]

def add_col(table:pyspark.sql.dataframe.DataFrame, col:str, default_value="", extra_client=None):
    """ Adds a column of empty values to an existing Spark Dataframe
    Args:
        table           :   Required (pyspark.sql.dataframe.DataFrame) - Spark Table
        col             :   Required (str) - Spark Table column name
        default         :   Optional (any) - Value to insert for every row in the newly created column
        extra_client    :   Ignore this value - necessary for other labelbase integrations
    Returns:
        Your table with a new column given the column_name and default_value  
    """
    table = table.withColumn(col, pyspark.sql.functions.lit(default_value))
    return table

def get_table_length(table:pyspark.sql.dataframe.DataFrame, extra_client=None):
    """ Tells you the size of a Spark Dataframe
    Args:
        table           :   Required (pyspark.sql.dataframe.DataFrame) - Spark Table
        extra_client    :   Ignore this value - necessary for other labelbase integrations
    Returns:
        The length of your table as an integer
    """  
    return table.count()

def rename_col(table:pyspark.sql.dataframe.DataFrame, col:str, to:str):
    """ Renames columns into the Labelspark format    
    Args:
        table           :   Required (pyspark.sql.dataframe.DataFrame) - Spark Table
        col             :   Required (str) - Existing column name to-be-changed
        to              :   Required (str) - What to rename the column as       
    Returns:
        Updated `table` object with renamed column
    """
    table = table.withColumnRenamed(col,to)
    return table                  
 



from pyspark.sql.functions import col
from labelspark.constants import LABELBOX_DEFAULT_TYPE_DICTIONARY

def dataframe_schema_enrichment(raw_dataframe, type_dictionary=None):
    if type_dictionary is None:
        type_dictionary = LABELBOX_DEFAULT_TYPE_DICTIONARY
    copy_dataframe = raw_dataframe
    for column_name in type_dictionary:
        try:
            copy_dataframe = copy_dataframe.withColumn(
                column_name,
                col(column_name).cast(type_dictionary[column_name]))
        except Exception as e:
            print(e.__class__, "occurred for", column_name, ":",
                  type_dictionary[column_name], ". Please check that column.")

    return copy_dataframe


from labelbox.schema.data_row_metadata import DataRowMetadataKind as lb_metadata_type
from pyspark import SparkContext
from packaging import version
import datetime
import json
from pyspark.sql.types import StructType, StructField, StringType, MapType, ArrayType
from pyspark.sql.functions import udf, lit

def create_labelbox_dataset(client, spark_dataframe, dataset_name=str(datetime.datetime.now()), add_data_row_ids=True, iam_integration='DEFAULT', metadata_index=False, **kwargs):
  """ Creates a Labelbox dataset and creates data rows given a spark dataframe. Uploads data rows in batches of 10,000.
  Args:
      client                  :     labelbox.Client object
      spark_dataframe         :     pyspark.sql.dataframe.Dataframe object - must have "row_data" and "external_id" columns at a minimum
      dataset_name            :     Labelbox dataset name      
      iam_integration         :     IAM integreation to use when creating the Labelbox dataset
      add_data_row_ids        :     Boolean - if True, adds a "data_row_id" column - if False, does nothing
      metadata_index          :     (Optional) Dictionary to add metadata to your data rows. Synatx is {key=column_name : value=metadata_type}
                                          Metadata type is one of the following strings:
                                              "enum"
                                              "string"
                                              "number"
                                              "datetime"
  Returns:
    spark_dataframe           :     updated spark_dataframe
    lb_dataset                :     labelbox.schema.dataset.Dataset object
  """
  lb_dataset = client.create_dataset(name=dataset_name, iam_integration=iam_integration, **kwargs)
  connect_spark_metadata(client, spark_dataframe, metadata_index) 
  print("Dataset created in Labelbox.")
  # Creates a column in dictionary format to-be-uploaded to Labelbox
  spark_dataframe = create_uploads_column(spark_dataframe, client, metadata_index)
  # Batch uploads data rows
  pandas_df = spark_dataframe.to_pandas_on_spark()
  print(f'Uploading {len(pandas_df)} to Labelbox in dataset with ID {lb_dataset.uid}')
  lb_dataset = batch_upload_data_rows(lb_dataset, pandas_df, 10000)
  # Adds a data_row_id column to your dataframe
  if add_data_row_ids:
    print(f'Attaching Laeblbox Data Row IDs to Spark Table')
    spark_dataframe = attach_data_row_ids(lb_dataset, spark_dataframe)
  return lb_dataset, spark_dataframe

### Connect Spark Metadata
def connect_spark_metadata(client, spark_dataframe, metadata_index):
  """ Checks to make sure all desired metadata for upload has a corresponding field in Labelbox. Note limits on metadata field options, here https://docs.labelbox.com/docs/limits
  Args:  
    client                    :    labelbox.Client object
    spark_dataframe           :    pyspark.sql.dataframe.Dataframe object
    metadata_index            :    Dictionary to add metadata to your data rows. Synatx is {key=column_name : value=either "enum", "string", "number" or "datetime"}
  Returns:
    Nothing - the metadata ontology has been updated
  """
  conversion = {
    "enum" : lb_metadata_type.enum,
    "string" : lb_metadata_type.string,
    "datetime" : lb_metadata_type.datetime,
    "number" : lb_metadata_type.number
  } 
  if metadata_index:
    print("Connecting Metadata to Labelbox")
    lb_metadata_index = {}      
    mdo = client.get_data_row_metadata_ontology()
    mdo_dict = mdo._get_ontology()
    labelbox_metadata_names = [field['name'] for field in mdo_dict]
    for column_name in metadata_index.keys():
      lb_metadata_index.update({column_name : conversion[metadata_index[column_name]]})  
    for spark_metadata_name in list(metadata_index.keys()):
      if spark_metadata_name not in labelbox_metadata_names:
        metadata_type = lb_metadata_index[spark_metadata_name]
        create_metadata_field(mdo, spark_dataframe, spark_metadata_name, metadata_type)
  if "lb_integration_source" not in labelbox_metadata_names:
      metadata_type = lb_metadata_type.string
      create_metadata_field(mdo, spark_dataframe, "lb_integration_source", metadata_type)
  return lb_metadata_index

### Connect Spark Metadata
def create_metadata_field(metadata_ontology_object, spark_dataframe, spark_metadata_name, metadata_type):
  """ Given a metadata field name and a column, creates a metadata field in Laeblbox given a labelbox metadata type
  Args:
    metadata_ontology_object  :    labelbox.schema.data_row_metadata.DataRowMetadataOntology object
    spark_dataframe           :    pyspark.sql.dataframe.Dataframe object
    spark_metadata_name       :    Name of the column in the pyspark table to-be-converted into a metadata field
    labelbox_metadata_type    :    labelbox.schema.data_row_metadata.DataRowMetadataKind object
  Returns:
    Nothing - the metadata ontology now has a new field can can be refreshed
  """
  if metadata_type == lb_metadata_type.enum:
    enum_options = [str(x.__getitem__(spark_metadata_name)) for x in spark_dataframe.select(spark_metadata_name).distinct().collect()]
  else:
    enum_options = None
  metadata_ontology_object.create_schema(name=spark_metadata_name, kind=metadata_type, options=enum_options) 
  
## Create Uploads Column
def create_uploads_column(spark_dataframe, client, metadata_index=False):
  """ Creates a colum using the pyspark StructType class that consists of row_data, external_id, and metadata_fields if the `metadata_index` argument is provided
  Args:
      client                  :     labelbox.Client object
      spark_dataframe         :     pyspark.sql.dataframe.Dataframe object - must have "row_data" and "external_id" columns at a minimum 
      metadata_index          :     (Optional) Dictionary where {key=column_name : value=metadata_type} (where metadata_type is "enum", "number", "string" or "datetime"). If not provided, will not add metadata to the data row
  Returns:
      spark_dataframe with an `uploads` column that can be converted to a Labelbox data row upload dictionary
  """
  # Grab the metadata ontology and create a dictionary where {key=metadata_name : value=metadata_feature_schema_id} or {key=metadata_name : value={"parent", "feature_schema_id"}} for enum metadata
  mdo = client.get_data_row_metadata_ontology()
  metadata_dict = mdo.reserved_by_name
  metadata_dict.update(mdo.custom_by_name)
  mdo_lookup = {}
  for name in metadata_dict:
    # Catches Enum Metadata
    if type(metadata_dict[name]) == dict:
      for enum_option in metadata_dict[name]:
        mdo_lookup.update({
          str(enum_option) : {
            "feature_schema_id" : metadata_dict[name][enum_option].uid, 
            "parent" : metadata_dict[name][enum_option].parent
          }
        })
    else:
      mdo_lookup.update({
        name : {
          "feature_schema_id" : metadata_dict[name].uid
        }
      })
  # Specify the structure of the `uploads` column
  upload_schema = StructType([
    StructField("row_data", StringType()),
    StructField("external_id", StringType()),
    StructField("metadata_fields", ArrayType(MapType(StringType(), StringType(), True)))
  ])
  # Create an `uploads` column with row_data and external_id
  create_uploads_udf = udf(create_uploads, upload_schema)
  df = spark_dataframe.withColumn('uploads', create_uploads_udf('row_data', 'external_id', lit(json.dumps(mdo_lookup))))
  # Attach metadata to the `uploads` column if metadata_index argument is provided
  if metadata_index:
    attach_metadata_udf = udf(attach_metadata, upload_schema)
    for column_name in metadata_index:
      df = df.withColumn('uploads', attach_metadata_udf(column_name, 'uploads', lit(column_name), lit(json.dumps(mdo_lookup)), lit(json.dumps(metadata_index))))
  return df

## Create Uploads Column
def create_uploads(row_data, external_id, mdo_lookup_bytes):
  """ Function to-be-wrapped into a user-defined function
  Args:
      row_data                :     row_data value
      external_id             :     external_id value      
      mdo_lookup_bytes        :     Bytearray representation of a dictionary where {key=metadata_name : value=metadata_feature_schema_id or value={"parent", "feature_schema_id"}} for enum metadata    
  Returns:
      Data row upload as-a-dictionary
  """
  mdo_lookup = json.loads(mdo_lookup_bytes)
  return {
    "row_data" : row_data,
    "external_id" : external_id,
    "metadata_fields" : [
      {
        "schema_id" : mdo_lookup["lb_integration_source"]['feature_schema_id'],
        "value" : "Databricks"
      }
    ]
  }

## Create Uploads Column
def attach_metadata(metadata_value, data_row, column_name, mdo_lookup_bytes, metadata_index_bytes):
  """ Function to-be-wrapped into a user-defined function
  Args:
      metadata_value          :     Value for the metadata field
      data_row                :     Data row dictionary to add metadata fields to
      column_name             :     Name of the column holding the metadata value
      mdo_lookup_bytes        :     Bytearray representation of a dictionary where {key=metadata_name : value=metadata_feature_schema_id or value={"parent", "feature_schema_id"}} for enum metadata    
      metadata_index_bytes    :     Bytearray representation of a dictionary where {key=metadata_name : value=metadata_type as string}
  Returns:
      Data row upload as-a-dictionary
  """  
  mdo_lookup = json.loads(mdo_lookup_bytes)
  metadata_type = json.loads(metadata_index_bytes)[column_name]
  if (metadata_type == "enum") and (metadata_value is not None):
    data_row['metadata_fields'].append({
      "schema_id" : mdo_lookup[str(metadata_value)]['parent'],
      "value" : mdo_lookup[str(metadata_value)]['feature_schema_id']     
    })
  elif (metadata_value is not None) or (str(metadata_value) != ""):
    data_row['metadata_fields'].append({
      "schema_id" : mdo_lookup[column_name]['feature_schema_id'],
      "value" : str(metadata_value)
    })
  return data_row  

# Batch Upload Data Rows
def batch_upload_data_rows(lb_dataset, pandas_df, upload_batch_size):
  """ Batch Uploads Data Row IDs given a list of uploads and a batch size
  Args: 
    pandas_df                 :     pyspark.pandas.frame.DataFrame version of the working pyspark dataframe
    upload_batch_size         :     Integer - Maximum is 30,000 if there's metadata on data rows
  Returns:
    updated Labelbox dataset DB Object
  """
  starttime = datetime.datetime.now()
  for i in range(0, len(pandas_df), upload_batch_size):
    if i+upload_batch_size<=len(pandas_df):
      batch_df = pandas_df.iloc[i:i+upload_batch_size]
    else:
      batch_df = pandas_df.iloc[i:]
    print(f'Batch Number {int(1+(i/upload_batch_size))} with {len(batch_df)} data rows')
    batch_spark_df = batch_df.to_spark()
    batch_upload = batch_spark_df.select("uploads").rdd.map(lambda x: x.uploads.asDict()).collect()
    task = lb_dataset.create_data_rows(batch_upload)
    task.wait_till_done() 
    print(f'Upload Speed: {datetime.datetime.now()-starttime}')
    print(f'Errors: {task.errors}')
    starttime = datetime.datetime.now()    
  return lb_dataset

# Attach data row IDs
def attach_data_row_ids(lb_dataset, spark_dataframe):
  """ UDF to add data row IDs to a dataframe given external_id and relevant labelbox dataset. Assumption is that external_id is unique
  Args:  
    spark_dataframe           :     pyspark.sql.dataframe.Dataframe object - must have "external_id" column
    lb_dataset                :     labelbox.schema.dataset.Dataset object
  Returns:
    pyspark dataframe with a completed data_row_id column filled in
  """
  external_id_to_data_row_id = {}
  data_rows = lb_dataset.export_data_rows()
  for data_row in data_rows:
    external_id_to_data_row_id.update({str(data_row.external_id) : str(data_row.uid)})
  find_data_row_id_udf = udf(find_data_row_id, StringType())
  return spark_dataframe.withColumn('data_row_id', find_data_row_id_udf("external_id", lit(json.dumps(external_id_to_data_row_id))))

# Attach data row IDs
def find_data_row_id(external_id, reference):
  """ Nested UDF Functionality to return a data row ID given the external ID
  Args:
    external_id               :      Data Row external ID as a string
    reference                 :      dictionary where {key=external_id : value=data_row_id}
  Returns:
    Data Row ID value
  """
  reference_dict = json.loads(reference)
  data_row_id = reference_dict[external_id]
  return data_row_id



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='labelspark',
    version='0.7.32',
    author='Labelbox',
    author_email="raphael@labelbox.com",      
    packages=find_packages(),
    url='https://github.com/Labelbox/LabelSpark.git',
    description='Labelbox Connector for Databricks',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["labelbox[data]", "pyspark", "packaging", "labelbase"],
    extras_require={'dev': ['pylint']}
)



'''
This code is derived  from tutorial "Working with Badly Nested Data in Spark" by by Paul Siegel
It produces a list of columns with periods between nested levels.

schema_fields_array = spark_schema_to_string(
        bronze_table.schema.jsonValue())

result looks like ['ColumnName.Category.Field1', 'ColumnName.Category.Field2',....]
'''
def spark_schema_to_string(schema, progress=''):
    if schema['type'] == 'struct':
        for field in schema['fields']:
            key = field['name']
            yield from spark_schema_to_string(field, f'{progress}.{key}')
    elif schema['type'] == 'array':
        if type(schema['elementType']) == dict:
            yield from spark_schema_to_string(schema['elementType'], progress)
        else:
            yield progress.strip('.')
    elif type(schema['type']) == dict:
        yield from spark_schema_to_string(schema['type'], progress)
    else:
        yield progress.strip('.')


import requests  # we need to handle a large frame-by-frame dataset for videos, so we use requests
import json

#this code block is needed for backwards compatibility with older Spark versions
from pyspark import SparkContext
from packaging import version
try:
  import pyspark.pandas as pd
except:
  import databricks.koalas as pd 
needs_koalas = False

from labelspark.jsonToDataFrame import jsonToDataFrame

def get_videoframe_annotations(bronze_video_labels, api_key, spark, sc):
    # This method takes in the bronze table from get_annotations and produces
    # an array of bronze dataframes containing frame labels for each project
    bronze_video_labels = bronze_video_labels.withColumnRenamed(
        "DataRow ID", "DataRowID")
    if needs_koalas:
        bronze = bronze_video_labels.to_koalas()
    else:
        bronze = bronze_video_labels.to_pandas_on_spark()

    # We manually build a string of frame responses to leverage our existing jsonToDataFrame code, which takes in JSON
    headers = {'Authorization': f"Bearer {api_key}"}
    master_array_of_json_arrays = []
    for index, row in bronze.iterrows():
        response = requests.get(row.Label.frames, headers=headers, stream=False)
        data = []
        for line in response.iter_lines():
            data.append({
                "DataRow ID": row.DataRowID,
                "Label": json.loads(line.decode('utf-8'))
            })
        massive_string_of_responses = json.dumps(data)
        master_array_of_json_arrays.append(massive_string_of_responses)

    array_of_bronze_dataframes = []
    for frameset in master_array_of_json_arrays:
        array_of_bronze_dataframes.append(jsonToDataFrame(frameset, spark, sc))

    return array_of_bronze_dataframes
